/*
 * Generic list manipulation (yet another...)
 *
 * The LIST_NAME_ macros are useful when elements are part of multiple lists at
 * the same time. Otherwise, use the LIST_* macros. Lists are implemented by
 * adding a prev/next pointer to the struct representing an item. Example:
 *
 *	struct blob
 *	{
 *		...
 *		LIST	(struct blob);
 *		...
 *	};
 *
 * To create an empty list:
 *
 *	struct blob *head = NULL;
 *
 * To iterate through all elements:
 *
 *	struct blob *p;
 *	for (p = head; p; p = LIST_NEXT(head, p))
 *		...
 *
 * Internally, the list is circular and that's why LIST_NEXT and LIST_PREV
 * need an additional pointer (to head or tail of the list respectively) in
 * order to know where to stop. This also means that any list element can be
 * used as the head or tail of the list. To add a (possibly uninitialized)
 * element to a list:
 *
 *	struct blob q;
 *	LIST_ADD(head, &q);
 *
 * To find the tail of the list (i.e. last element added):
 *
 *	tail = LIST_TAIL(head);
 *
 * To remove an element from the list:
 *
 *	LIST_CUT(head, p);
 *
 * `p' will now be a list on its own with only one element. `head' will be set
 * to NULL when there are no elements left anymore (i.e. when head == p and
 * no next). To arbitrarily delete malloc'ed items from a list:
 *
 *	struct blob *p, *next;
 *	for (p = head; p; p = next)
 *	{
 *		next = LIST_NEXT(head, p);	// first obtain the next
 *		if (...)
 *		{
 *			LIST_CUT(head, p);	// and cut `p' from the list
 *			free(p);		// now `p' may be freed
 *		}
 *	}
 *
 * To delete an entire malloc'ed list:
 *
 *	struct blob *p;
 *	while ((p = head))
 *	{
 *		LIST_CUT(head, p);
 *		free(p);
 *	}
 *
 * Interface notes:
 * -	Every element (head, tail, elem below) is a list on its own and can
 *	therefor be NULL, meaning the empty list.
 * -	`item' cannot be NULL.
 */

#define LIST(type)		LIST_NAME(type, list)
#define LIST_TAIL(head)		LIST_NAME_TAIL(head, list)
#define LIST_NEXT(head, elem)	LIST_NAME_NEXT(head, list, elem)
#define LIST_PREV(tail, elem)	LIST_NAME_PREV(tail, list, elem)
#define LIST_ADD(head, item)	LIST_NAME_ADD(head, list, item)
#define LIST_CUT(head, item)	LIST_NAME_CUT(head, list, item)

#define LIST_NAME(type, name)							\
	struct { type *next, *prev; } _##name

#define LIST_NAME_TAIL(head, name)						\
	LIST_NAME_PREV(NULL, name, head)

#define LIST_NAME_NEXT(head, name, elem)					\
	(elem) ? ((elem)->_##name.next == (head) ? NULL : (elem)->_##name.next)	\
	     : (head)

#define LIST_NAME_PREV(tail, name, elem)					\
	(elem) ? ((elem)->_##name.prev == (tail) ? NULL : (elem)->_##name.prev)	\
	     : (tail)

#define LIST_NAME_ADD(head, name, item)						\
do {										\
	if (head)								\
	{									\
		(item)->_##name.next = (head);					\
		(item)->_##name.prev = (head)->_##name.prev;			\
		(head)->_##name.prev = (item);					\
		(item)->_##name.prev->_##name.next = (item);			\
	}									\
	else									\
	{									\
		(head) = (item);						\
		(head)->_##name.next = (head)->_##name.prev = (head);		\
	}									\
} while (0)

#define LIST_NAME_CUT(head, name, item)						\
do {										\
	if ((item) == (head))							\
	{									\
		(head) = (head)->_##name.next;					\
		if ((item) == (head))						\
		{								\
			(head) = NULL;						\
		}								\
	}									\
	(item)->_##name.next->_##name.prev = (item)->_##name.prev;		\
	(item)->_##name.prev->_##name.next = (item)->_##name.next;		\
	(item)->_##name.next = (item)->_##name.prev = (item);			\
} while (0)
